function Params = clusterDataMfa(X,Lh,Ph,Mu,Pi)
% Calculates the labels for a data set under a mixture of factor
% analysis model.
%
% X - data matrix
% Lh - factor loadings 
% Ph - diagonal uniquenesses matrix
% Mu - mean vectors
% Pi - priors
%
% Params:
%    {1}: Lbls - cluster labels. The number is determined by the number of means (Mu)
%    {2}: Mu - means of the clusters 
%    {3}: MM - covariance matrices
%    {4}: the mahalanobis distance from every element to the center of its
%    cluster
    N=length(X(:,1));
    D=length(X(1,:));
    K=length(Lh(1,:));
    M=length(Pi);
    if (abs(sum(Pi)-1) > 1e-6) 
        disp('ERROR: Pi should sum to 1');
        return;
    elseif ((size(Lh) ~= [D*M K]) | (size(Ph) ~= [D 1]) | (size(Mu) ~= [M D]) ...
        | (size(Pi) ~= [M 1] & size(Pi) ~= [1 M]))   
        disp('ERROR in input matrix sizes');
        return;
    end;  

    tiny=exp(-744);
    const=(2*pi)^(-D/2);

    I=eye(K);
    Phi=1./Ph;
    Phid=diag(Phi);
    Params = cell(4,1);
    CovMatrices = cell(M,1);
    for k=1:M  
        Lht=Lh((k-1)*D+1:k*D,:);
        LP=Phid*Lht;
        MM=Phid-LP*inv(I+Lht'*LP)*LP';
        dM=sqrt(det(MM));      	
        Xk=(X-ones(N,1)*Mu(k,:)); 
        XM=Xk*MM; 
        H(:,k)=const*Pi(k)*dM*exp(-0.5*sum((XM.*Xk)'))'; 	
        
        covMatr = Phid + Lht*Lht';
        CovMatrices{k} = Phid;%MM;%covMatr;
     end;
     [C, Lbls] = max(H, [], 2);
     Params{1} = Lbls;
     Params{2} = Mu;
     Params{3} = CovMatrices;
     weights_nu = computeWeights(X, Lbls);
     Params{4} = weights_nu{1};
     Params{5} = weights_nu{2} / N; % nu is set to the estimated number of outliers
     % substitute X with the dimensionality reduced data
     %Params{6} = computeProjections(X, Lbls, Lh, Ph, Mu);
     
function weights_nu = computeWeights(X, Lbls)     
    uniqLbls = unique(Lbls);
    weights = zeros(size(X,1),1);
    nu_estimate = 0;
    for i = 1:size(uniqLbls,1)
        I = find(Lbls == uniqLbls(i,1));
        Xi = X(I,:);
        if (size(Xi,1) <= size(Xi,2))
            W = ones(size(Xi,1), 1)*0.00001;
        else 
            W = mahal(Xi,Xi);
        end;
        weights(I,1) = W;
        m = mean(W);
        s = 1.5*std(W);
        outliers = find(W>(m+s));
        nu_estimate = nu_estimate+size(outliers,1);
    end;
    weights_nu{1} = weights;
    weights_nu{2} = nu_estimate;

function Z = computeProjections(X, Lbls, Lh, Ph, Mu)
    N=length(X(:,1));
    D=length(X(1,:));
    K=size(Mu,1);
    beta = [];
    % compute the beta_j vectors. See "The EM Algorithm For Mixture Of
    % Factor Analyzers" (Ghahramani)
    Ph = diag(Ph);
    for i = 1:K
        lambda_j = Lh(1+(i-1)*D:i*D,:);
        beta_j = lambda_j'*inv(Ph+lambda_j*lambda_j');
        beta = [beta; beta_j];        
    end;
    Z = [];
    size(beta)
    for i = 1:size(Lbls,1)
        lbl = Lbls(i);
        startInd = 1+(lbl-1)*D
        endInd = lbl*D
        beta_i = beta(startInd:endInd,:);
        mu_i = Mu(lbl,:);
        z_i = beta_i*(X(i,:) - mu_i)';
        Z = [Z;z_i'];
    end;
    figure;
    plot(Z(:,1), Z(:,2), '.b', 'LineWidth', 2);
