% Dragomir Yankov
% The same as ConstrainedSVC.m, but invokes the matlab qprog solver instead
% of libsvm, and adds weights to the examples based to their Mahalanobis
% distance to the center of their cluster computed by MFA
%
% Constrained Support Vector Clustering:
% Use the clusters generated from a mixture of factor analyzers (MFA)
% to provide better constraints (or kernel matrix) for the one class
% classification (extended to support vector clustering)
function ConstrainedSVC_withWeights(X, Y, file, num_clusters)
    if (nargin == 0)
        close all;
        clear all;
        file = 'data_unifang_radialgauss1.data';
        %file = 'dataSwissRoll_noise.data';
        data_dimensions = 2;
        [X, Y] = loadDataXY(file, data_dimensions);
    end;
    plotOrigData(X,Y);
    tic
    % run the mixture of factor analyzers and get the 
    mixComponents = 10;%round(size(X,1)/10);  % number of gaussian components 
    lowerDimension = 2;  % lower dimension k of the factor analyzers
    plotting_on = 1;
    Params = getMfaParams(X, mixComponents,lowerDimension);
    if (plotting_on && size(X,2) == 2)
        plotMfa(X, Params);
    end;
    
    % run support vector clustering constraining with the parameters
    % obtained by mfa
    if (nargin < 4)
        clusters = 2;%size(unique(Y),1); % assuming that we know the number of clusters as in K-means  
    else
        clusters = num_clusters;
    end;
    
    
    use_weights = 1 % 0 - not using weights. This is equivalent to the standard SVC
                    % 1 - using weights, this is the locally constrained SVC    
    merge_method = 2; % 0 merge multiple clusters to obtain the required number of clusters using the closest pairwise distance
                      % 1 merge the clusters using the closest mfa centers 
                      % 2 merge to the largest clusters
    %nu = 0.08; %not used. Computed from the clusters
    if (use_weights == 1)
        nu = Params{5};
    else
        nu = 1;
    end;
    gamma = 6;  % if gamma==0 estimate the optimal value, else use this value  
    [quality, labeling] = runSVC(X, Params, nu, gamma, Y, clusters, merge_method, use_weights, file);
    ClustAccuracy = quality
    toc
    
function [quality, labeling] = runSVC(X, Params, nu, gamma, Y, clusters, merge_method, use_weights, file)
    % compute kernel matrix, measuring the distance to the mfa cluster
    % centers
    Lbls = Params{1};
    clustMeans = Params{2};
    weights = Params{4};
    if (~use_weights)
        weights = ones(size(X,1),1);
    end;
    X_ = X;
    
    % run the constrained support vector clustering using the computed with MFA weights
    if (gamma == 0)
        SvmParams = 0;
        %for i = -8:0.5:8;
        fid = fopen([file '.bestgamma' num2str(use_weights)],'w');
        for i = -3:1:8
            tmpgamma = 2^(i);
            plotting = false;
            tmpSvmParams = runMatSvm(X_, nu, tmpgamma, weights, clustMeans, plotting);            
            if (~iscell(SvmParams) || betterParams(tmpSvmParams, SvmParams, clusters))
                SvmParams = tmpSvmParams;
                gamma = tmpgamma;
                fprintf(fid,'%8.4f\n',gamma);                
            end;
        end;
        plotting = true;
        % call this again just because of the contour plot
        SvmParams = runMatSvm(X_, nu, gamma, weights, clustMeans, plotting);
        fprintf(fid,'%8.4f\n',gamma);
        fclose(fid)
    else
        plotting = true;
        SvmParams = runMatSvm(X_, nu, gamma, weights, clustMeans, plotting);
    end;
  %  plotClustering(X_, SvmParams{1}, SvmParams{2}, SvmParams{3}, 0);                
    assign_bsv = 1; % 0 - don't assing the bounded support vectors to any cluster
                    % 1 - assign the bounded support vectors to the closest
                    % cluster
    clustLbls = mergeClusters(X_, SvmParams{3}, clusters, merge_method, Params, SvmParams{4}, assign_bsv);
    %compute the accuracy of the clustering
    if (size(unique(Y),1) > 1)
   %     [quality, labeling, permut] = testClusterQuality(clusters, Y, clustLbls);    
        quality = 1;
        labeling = clustLbls;
    else
        quality = 0;
        labeling = clustLbls;
    end;
    %plot the final cluster assignment
    plotClustering(X_, SvmParams{1}, SvmParams{2}, labeling, 1);
    bsvs = find(SvmParams{3} == 0);
    cls = unique(SvmParams{3});
    num_bsvs = size(bsvs,1)
    num_clusters = size(cls,1)

    
function Params = getMfaParams(X, mixComponents, lowerDimension)    
    [Lh,Ph,Mu,Pi,LL]=mfa(X,mixComponents,lowerDimension);
    Params = clusterDataMfa(X, Lh, Ph, Mu, Pi); % returns a cell array. The first element the labels, the second - the cluster means, the third - the covariances
    
function Ind = getIndices(X, svs)
    Ind = [];
    for i = 1:size(svs,1)
        tmpInd = findInd(X, svs(i,:));
        Ind = [Ind; tmpInd];
    end;
    
function Ind = findInd(X, sv)
    Ind = [];
    for i = 1:size(X,1)
        isInd = 1;
        for j = 1:size(X,2)
            if (X(i,j) ~= sv(1,j))
                isInd = 0;
                break;
            end;
        end;
        if (isInd == 1)
            Ind = i;
            return;
        end;
    end;
    
% compare two values of gamma. Choose the value for which the largest
% numClusters include more elements
function b = betterParams(newSvmParams, oldSvmParams, numClusters)   
    merger = manifolds.MergeClusters;
    C = oldSvmParams{3};
    size(C)
    oldlargest_cls = merger.largestClusters(oldSvmParams{3}, numClusters);
    % SVC did not detect enough clusters with the previous parameters,
    % whatever the new parameters are accept them
    if (oldlargest_cls(numClusters,1) == 0) 
        b = true;
        return;
    end;
    newlargest_cls = merger.largestClusters(newSvmParams{3}, numClusters);
    if (newlargest_cls(numClusters,1) == 0) 
        b = false;
        return;
    end;
    %find the number of elements in the maximal clusters for the old gamma
    numOldElements = 0;
    for i = 1:numClusters
        numOldElements = numOldElements + size(find(oldSvmParams{3} == oldlargest_cls(i,1)),1);
    end;
    %find the number of elements in the maximal clusters for the new gamma
    numNewElements = 0;
    newLbls = newSvmParams{3};
    newEls = [];
    for i = 1:numClusters
        tmpnewEls = size(find(newSvmParams{3} == newlargest_cls(i,1)),1);
        newEls = [newEls; tmpnewEls];
        numNewElements = numNewElements + tmpnewEls;
    end;
    % if the smallest cluster is too small stay with the old parameters
    if (min(newEls)/max(newEls) < 0.05)
        b= false;
        return;
    end;
    
    % the parameters that lead to larger clusters are selected as better
    b = (numNewElements > numOldElements);
    
function [X, Y] = loadDataXY(file, dims)
    dims_file = dims+1;
    D = loadData(file,dims_file);
    cols= size(D,2);
    X = D(:,1:cols-1);
    Y = D(:,cols);
    
    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Plotting functions
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function plotMfa(X, Params)
    figure;
    Lbls = Params{1};
    ClusterMeans = Params{2};
    CovMatrices = Params{3};
    lbls_type = unique(Lbls);
    num_lbls = size(lbls_type,1);
    %color_map = ['+b'; '*r'; 'xg'; 'om'; '^y'; 'vc'];
    color_map1 = [0 0 1; 1 0 0; 128/256 128/256 255/256; 10/256 164/256 10/256; 1 0 1; 0 1 1];
    color_symb = ['+'; '*'; 'x'; 'o'; '^'; 'v']; 
    hold on;
    for i = 1:num_lbls
        inds = find(Lbls(:,1) == lbls_type(i));
        col = rem(i-1, size(color_map1,1))+1;
        ncol = color_map1(col,:); 
        scol = color_symb(col,:);
        plot(X(inds,1),X(inds,2), scol, 'MarkerEdgeColor', ncol, 'LineWidth', 2);
        meanCluster = ClusterMeans(lbls_type(i),:);
        examp = X(inds,:);
        covMatr = cov(examp);
        size(examp)
        if size(examp,1) < 2
            continue;
        end;
        draw_ellipse(meanCluster, covMatr, ncol, 2);  
        hold on;
    end;
    axis off;
    %axis([-4,4,-4,4]); 
    
function plotOrigData(X, Y)    
    inds = find(Y == 1);
    figure;
    lbls_type = unique(Y);
    num_lbls = size(lbls_type,1);
    %color_map = ['+b'; '*r'; 'xg'; 'om'; '^y'; 'vc']; 
    color_map1 = [0 0 1; 1 0 0; 128/256 128/256 255/256; 10/256 164/256 10/256; 1 0 1; 0 1 1];
    color_symb = ['+'; '*'; 'x'; 'o'; '^'; 'v']; 
    for i = 1:num_lbls
        inds = find(Y(:,1) == lbls_type(i,1));
        if (lbls_type(i,1) == 0) % plot the BSVs
            scol = 'd';
            ncol = [0 0 0];
        else 
            col = rem(i-1, size(color_map1,1))+1;
            scol = color_symb(col,:);
            ncol = color_map1(col,:);
        end;
        if (size(X,2) == 2)
            plot(X(inds,1),X(inds,2), scol, 'MarkerEdgeColor', ncol, 'LineWidth', 3);        
        else
            plot3(X(inds,1),X(inds,2), X(inds,3), scol, 'MarkerEdgeColor', ncol, 'LineWidth', 2);        
        end;
        hold on;            
    end;
    box off
    set(gca,'XtickLabel','');   
    set(gca,'YtickLabel','');   
 %   axis off;
    
function C = getColormap()
    X = [0.6 0.2 0.39]
    Y = [0.601 0.601 1]
    Z = [0.746 0.746 0.601];
    step1 = (Y-X)/3;
    step2 = (Z-Y)/3;
    C = [X;X+step1;X+2*step1;Y;Y+step2;Y+2*step2;Z];
    
function plotClustering(X, model, train_label, Lbls, newfigure)
    svs = full(model.SVs);
    sv_inds = getIndices(X, svs);
    inds_neg = find(train_label < 0);
    inds_pos = find(train_label > 1);
    if (size(X,2) > 2 || newfigure)
  %      figure;
    end;
    lbls_type = unique(Lbls);
    num_lbls = size(lbls_type,1);
    color_map1 = [0 0 1; 1 0 0; 128/256 128/256 255/256; 10/256 164/256 10/256; 1 0 1; 0 1 1];
    %color_map = ['+b'; '*r'; 'xg'; 'om'; '^y'; 'vc']; 
    color_symb = ['+'; '*'; 'x'; '.'; '^'; 'v']; 
    %color_symb = ['+'; '+'; '+'; '+'; '+'; '+']; 
    for i = 1:num_lbls
        inds = find(Lbls(:,1) == lbls_type(i,1));
%        if (lbls_type(i,1) == 0) % plot the BSVs
%            scol = 'd';
%            ncol = [0 0 0];
%        else 
            col = rem(i-1, size(color_map1,1))+1;
            scol = color_symb(col,:);
            ncol = color_map1(col,:);
            %        end;
        if (size(X,2) == 2)
            plot(X(inds,1),X(inds,2), scol, 'MarkerEdgeColor', ncol, 'LineWidth', 2);        
        else
            plot3(X(inds,1),X(inds,2), X(inds,3), scol, 'MarkerEdgeColor', ncol, 'LineWidth', 2);        
        end;
        hold on;            
    end;
    %xlabel('X');
    %ylabel('Y');
    %zlabel('Z');
    if (size(X,2) == 2)
        plot(svs(:,1), svs(:,2), 'ok', 'LineWidth', 2);     
    else
        %plot3(svs(:,1), svs(:,2), svs(:,3), 'ok', 'LineWidth', 2);     
    end;
    box off
    set(gca, 'FontSize', 14);
    %set(gca,'XtickLabel','');   
    %set(gca,'YtickLabel','');   
    %axis([-4,4,-4,4]);    
    %axis([-2,2,-2,2]);